/*
 * Filename: directedpathopening.c
 *
 * Does constrained and unconstrained path openings and closings.
 *
 * Author:
 * Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * Copyright 2008-2009.
 *
 * This and accompanying files contain code exactly as it is in DIPlib
 * for the functions dip_DirectedPathOpening() and dip_PathOpening().
 * Here we modified only the function names to avoid name conflicts
 * when linking against DIPlib.
 *
 * This code may be used freely for research purposes only. Do not
 * copy or distribute without the prior written consent by the author.
 *
 */

/*
 * To compile, type on the MATLAB command prompt:
 *
 *   dipmex directedpathopening.c pathopening_commoncode.c -I.
 *
 * Note you need to have DIPimage installed and initialized for this to work.
 * You must also have configured the MEX command by running:
 *
 *   mex -setup
 */

/*
 * See:
 *    H. Heijmans, M. Buckley and H. Talbot, "Path Openings and Closings",
 *       Journal of Mathematical Imaging and Vision 22:107-119, 2005.
 *    H. Talbot, B. Appleton, "Efficient complete and incomplete path
 *       openings and closings", Image and Vision Computing 25:416-425, 2007.
 *
 * Basic idea: opening with a linear structuring element that is flexible.
 * The line can adapt to local image features. It takes steps within a 90
 * degree cone of the chosen direction. Thus, if the chosen direction is
 * horizontal, it can take steps in the directions [1,-1], [1,0] and [1,1].
 *
 * Modified thus: only one consecutive not-straight step is allowed. The
 * straight step is the one in the chosen direction. If the chosen direction
 * is horizontal ([1,0]), you can't follow a [1,-1] step by another [1,-1]
 * step nor by a [1,1] step, only by a [1,0] step. But a [1,0] step can
 * be followed by any of the 3 allowed steps. This avoids creating zig-zag
 * lines, which are longer than you want. A thick line in the image can
 * fit a longer line (if it zig-zags) than a 1-pixel thin line of the same
 * length. By reducing the zig-zagging, this bias is also reduced.
 *
 * I modified the Talbot & Appleton algorithm by creating two "length" images
 * for the upstream direction and two for the downstream direction. The one
 * image holds the "straight" length, the other one holds the length as in
 * the original algorithm. By propagating the lengths to the right length
 * image, we can accomplish our goal.
 *
 * Definition of filterParam:
 * length = max(abs(filterParam)) is the number of pixels in the line.
 * The angle can be either 0, 45, 90 or 135 degrees in 2D, and similar stuff
 * in 3D. To decode this value, the filterParam is translated to an array with
 * -1, 0 and 1 values using:
 *    direction = round(filterParam/length)
 * direction==[1,0] means steps always go horizontally, and thus [1,1], [1,0]
 * and [1,-1] are valid steps for the line to take. [1,1] means steps are in
 * the +45 degree direction, meaning [1,0], [1,1] and [0,1] are valid steps to
 * take.
 * This generalizes easily to any number of dimensions.
 *
 */

#include "dml_dipmex.h"         /* includes diplib.h and mex.h */
#include "dip_developer.h"      /* for dip__ImageSetStride()*/
#include "dip_manipulation.h"   /* for dip_SetInteger() */
#include "dip_sort.h"           /* for dip_SortIndices() */
#include "dip_regions.h"        /* for dip_PlaneDoEdge() */
#include "dip_pixelqueue.h"     /* for dip_PixelQueue */

#include "pathopening.h"

dip_Error DCPY_DirectedPathOpening
(
   dip_Image in,
   dip_Image mask,
   dip_Image p_out,
   dip_FloatArray filterParam,
   dip_Boolean closing,        /* Set to DIP_TRUE to do the area closing instead */
   dip_Boolean constrained     /* Set to DIP_TRUE to use the constrained version, DIP_FALSE the original Talbot & Appleton version */
)
{
   DIP_FNR_DECLARE("DCPY_DirectedPathOpening");
   dip_Image out, active, len1, len2, len3, len4;
   dip_IntegerArray direction;
   dip_int length, ii, ndims, nneigh;
   dip_IntegerArray stride, imsz, indxup, indxdn;
   dip_DataType ovlDataType;
   dip_ImageArray imar, omar, nwar;
   dip_VoidPointerArray vpar;
   dip_int *indices, noInd;
   DIP_FNR_INITIALISE;

   /* Check inputs */
   DIPXJ( dip_IsScalar( in, 0 ));
   DIPXJ( dip_ImageGetDataType( in, &ovlDataType ));
   DIPXJ( dip_DataTypeAllowed( ovlDataType, DIP_TRUE, DIP_DT_IS_REAL, 0 ));
   if( mask ) {
      DIPXJ( dip_ImagesCompareTwo( in, mask, DIP_CPIM_SIZE_MATCH | DIP_CPIM_TYPES_MATCH, 0 ));
   }

   DIPXJ( dip_ImageGetDimensions( in, &imsz, rg ));
   ndims = imsz->size;
   DIPTS( ndims < 2, DIP_E_DIMENSIONALITY_NOT_SUPPORTED );
   for( ii=0; ii<ndims; ii++ ) {
      DIPTS( imsz->array[ii] < 3, "Input image is too small." );
   }
   DIPTS( filterParam->size != ndims, "Image and filter size dimensionalities do not match." );
   length = (dip_int)DIP_ABS( filterParam->array[ 0 ] );
   for( ii=1; ii<ndims; ii++ ) {
      dip_int tmp = (dip_int)DIP_ABS( filterParam->array[ ii ] );
      if( tmp > length ) {
         length = tmp;
      }
   }
   DIPTS( length < 2 || length > DIP_MAX_UINT16, "Illegal length value." );
   DIPXJ( dip_IntegerArrayNew( &direction, ndims, 0, rg ));
   for( ii=0; ii<ndims; ii++ ) {
      direction->array[ ii ] = dipm_Round( filterParam->array[ ii ] / length );
   }

   /* Prepare output image */
   if( p_out == in ) {
      /* Allow in-place operation */
      out = in;
   }
   else {
      DIPXJ( dip_ImageArrayNew( &imar, 2, rg));
      DIPXJ( dip_ImageArrayNew( &omar, 1, rg));
      imar->array[0] = in;
      imar->array[1] = mask;
      omar->array[0] = p_out;
      DIPXJ( dip_ImagesSeparate( imar, omar, &nwar, 0, rg ));
      out = nwar->array[0];
      DIPXJ( dip_Copy( in, out ));
   }
   DIPXJ( dip_ImageGetStride( out, &stride, rg ));

   /* Prepare temporary images */
   DIPXJ( dip_ImageNew( &active, rg ));
   DIPXJ( dip_ImageCopyProperties( out, active ));
   DIPXJ( dip__ImageSetStride( active, stride ));
   DIPXJ( dip_ImageSetDataType( active, DIP_DT_BIN8 ));
   DIPXJ( dip_ImageForge( active ));
   if( mask ) {
      DIPXJ( dip_Copy( mask, active ));
   }
   else {
      DIPXJ( dip_SetInteger( active, DIP__PO_ACTIVE, 0, DIP_FALSE ));
   }
   DIPXJ( dip_PlaneDoEdge( active, 0, 0 )); /* Set border pixels to inactive, we won't process them. */

   DIPXJ( dip_ImageNew( &len1, rg ));
   DIPXJ( dip_ImageCopyProperties( out, len1 ));
   DIPXJ( dip__ImageSetStride( len1, stride ));
   DIPXJ( dip_ImageSetDataType( len1, DIP_DT_UINT16 ));
   DIPXJ( dip_ImageForge( len1 ));
   DIPXJ( dip_SetInteger ( len1, length, 0, DIP_FALSE ));

   DIPXJ( dip_ImageNew( &len2, rg ));
   DIPXJ( dip_ImageCopyProperties( out, len2 ));
   DIPXJ( dip__ImageSetStride( len2, stride ));
   DIPXJ( dip_ImageSetDataType( len2, DIP_DT_UINT16 ));
   DIPXJ( dip_ImageForge( len2 ));
   DIPXJ( dip_SetInteger ( len2, length, 0, DIP_FALSE ));

   if( constrained ) {

      DIPXJ( dip_ImageNew( &len3, rg ));
      DIPXJ( dip_ImageCopyProperties( out, len3 ));
      DIPXJ( dip__ImageSetStride( len3, stride ));
      DIPXJ( dip_ImageSetDataType( len3, DIP_DT_UINT16 ));
      DIPXJ( dip_ImageForge( len3 ));
      DIPXJ( dip_SetInteger ( len3, length, 0, DIP_FALSE ));

      DIPXJ( dip_ImageNew( &len4, rg ));
      DIPXJ( dip_ImageCopyProperties( out, len4 ));
      DIPXJ( dip__ImageSetStride( len4, stride ));
      DIPXJ( dip_ImageSetDataType( len4, DIP_DT_UINT16 ));
      DIPXJ( dip_ImageForge( len4 ));
      DIPXJ( dip_SetInteger ( len4, length, 0, DIP_FALSE ));

   }
   else {
      len3 = len4 = 0;
   }

   /* Get data pointers */
   DIPXJ( dip_ImageArrayNew( &imar, 6, rg));
   imar->array[0] = out;
   imar->array[1] = active;
   imar->array[2] = len1;
   imar->array[3] = len2;
   imar->array[4] = len3;
   imar->array[5] = len4;
   DIPXJ( dip_ImageGetData( imar, &vpar, 0, 0, 0, 0, 0, rg ));

   /* Create sorted indices array (skipping border) */
   if( mask ) {
      DIPXJ( DCPY__IndicesArrayCreateMask( mask, stride, &indices, &noInd, rg ));
   }
   else {
      DIPXJ( DCPY__IndicesArrayCreate( imsz, stride, &indices, &noInd, rg ));
   }
   DIPXJ( dip_SortIndices( vpar->array[ 0 ], indices, noInd, DIP_SORT_DEFAULT,
                           ovlDataType, DIP_DT_INT ));
   if( closing ) {
      /* Just invert the sorted indices -- why isn't this an option in dip_Sort ??? */
      dip_int tmp;
      for( ii=0; ii<noInd/2; ii++ ) {
         tmp = indices[ii];
         indices[ii] = indices[noInd-ii-1];
         indices[noInd-ii-1] = tmp;
      }
   }

   /* Create two arrays with indices to neighbours */
   nneigh = dipm_PowInt( 3, ndims-1 ); /* number of upstream neighbours identical to number downstream. */
   DIPXJ( dip_IntegerArrayNew( &indxup, nneigh, 0, rg ));
   DIPXJ( dip_IntegerArrayNew( &indxdn, nneigh, 0, rg ));
   DIPXJ( DCPY__MakeNeighbourLists( direction, stride, indxup, indxdn ));

   /* Do the data-type-dependent thing */
   if( constrained ) {
#define DIP_OVL_ALLOW DIP_DTGID_REAL
#define DIP_OVL_FUNC DCPY__ConstrainedPathOpening
#define DIP_OVL_ARGS ( vpar->array[ 0 ], vpar->array[ 1 ], vpar->array[ 2 ], vpar->array[ 3 ], \
                       vpar->array[ 4 ], vpar->array[ 5 ], indices, noInd, indxup, indxdn, length )
#include "dip_ovl.h"
   }
   else {
#define DIP_OVL_ALLOW DIP_DTGID_REAL
#define DIP_OVL_FUNC DCPY__PathOpening
#define DIP_OVL_ARGS ( vpar->array[ 0 ], vpar->array[ 1 ], vpar->array[ 2 ], vpar->array[ 3 ], \
                       indices, noInd, indxup, indxdn, length )
#include "dip_ovl.h"
   }

dip_error:
   DIP_FNR_EXIT;
}


void mexFunction (int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]) {
   dip_Image in;
   dip_Image mask;
   dip_Image out;
   dip_FloatArray filterParam;
   dip_Boolean closing;
   dip_Boolean constrained;
   DML_ERROR_INIT;
   DML_CHK_NARGSIN (5);
   DML_CHK_NARGSOUT (1);
   DML_2DIP_IMAGE (0, in);
   DML_2DIP_MASKIMAGE (1, mask);
   DML_GENDIP_IMAGE (out);
   DML_2DIP_FLOATARRAY (2, filterParam);
   DML_2DIP_BOOLEAN (3, closing);
   DML_2DIP_BOOLEAN (4, constrained);
   DMLXJ ( DCPY_DirectedPathOpening (in, mask, out, filterParam, closing, constrained) );
   DML_2MEX_IMAGE (0, out);
dml_error:
   DML_ERROR_EXIT;
}
