/*
 * Filename: pathopening.c
 *
 * Does directedpathopening.c for each possible direction.
 *
 * Author:
 * Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * Copyright 2008-2009.
 *
 * This and accompanying files contain code exactly as it is in DIPlib
 * for the functions dip_DirectedPathOpening() and dip_PathOpening().
 * Here we modified only the function names to avoid name conflicts
 * when linking against DIPlib.
 *
 * This code may be used freely for research purposes only. Do not
 * copy or distribute without the prior written consent by the author.
 *
 */

/*
 * To compile, type on the MATLAB command prompt:
 *
 *   dipmex pathopening.c pathopening_commoncode.c -I.
 *
 * Note you need to have DIPimage installed and initialized for this to work.
 * You must also have configured the MEX command by running:
 *
 *   mex -setup
 */

#include "dml_dipmex.h"         /* includes diplib.h and mex.h */
#include "dip_developer.h"      /* for dip__ImageSetStride()�*/
#include "dip_manipulation.h"   /* for dip_SetInteger() */
#include "dip_sort.h"           /* for dip_SortIndices() */
#include "dip_regions.h"        /* for dip_PlaneDoEdge() */
#include "dip_math.h"           /* for dip_Max() */
#include "dip_pixelqueue.h"     /* for dip_PixelQueue */

#include "pathopening.h"

dip_Error DCPY_PathOpening
(
   dip_Image in,
   dip_Image mask,
   dip_Image p_out,
   dip_int length,
   dip_Boolean closing,        /* Set to DIP_TRUE to do the area closing instead */
   dip_Boolean constrained     /* Set to DIP_TRUE to use the constrained version, DIP_FALSE the original Talbot & Appleton version */
)
{
   DIP_FNR_DECLARE("DCPY_PathOpening");
   dip_Image out, tmp, active, len1, len2, len3, len4;
   dip_IntegerArray direction;
   dip_int ii, nneigh, ndims;
   dip_Boolean valid, firstone;
   dip_IntegerArray stride, imsz, indxup, indxdn;
   dip_DataType ovlDataType;
   dip_ImageArray imar, omar, nwar;
   dip_VoidPointerArray vpar;
   dip_int *indices, noInd;
   DIP_FNR_INITIALISE;

   /* Check inputs */
   DIPXJ( dip_IsScalar( in, 0 ));
   DIPXJ( dip_ImageGetDataType( in, &ovlDataType ));
   DIPXJ( dip_DataTypeAllowed( ovlDataType, DIP_TRUE, DIP_DT_IS_REAL, 0 ));
   if( mask ) {
      DIPXJ( dip_ImagesCompareTwo( in, mask, DIP_CPIM_SIZE_MATCH | DIP_CPIM_TYPES_MATCH, 0 ));
   }

   DIPXJ( dip_ImageGetDimensions( in, &imsz, rg ));
   ndims = imsz->size;
   DIPTS( ndims < 2, DIP_E_DIMENSIONALITY_NOT_SUPPORTED );
   for( ii=0; ii<ndims; ii++ ) {
      DIPTS( imsz->array[ii] < 3, "Input image is too small." );
   }
   DIPTS( length < 2 || length > DIP_MAX_UINT16, "Illegal length value." );

   /* Prepare output image */
   DIPXJ( dip_ImageArrayNew( &imar, 2, rg));
   DIPXJ( dip_ImageArrayNew( &omar, 1, rg));
   imar->array[0] = in;
   imar->array[1] = mask;
   omar->array[0] = p_out;
   DIPXJ( dip_ImagesSeparate( imar, omar, &nwar, 0, rg ));
   out = nwar->array[0];

   /* Prepare temporary images */
   DIPXJ( dip_ImageNew( &tmp, rg ));
   DIPXJ( dip_Copy( in, tmp ));
   DIPXJ( dip_ImageGetStride( tmp, &stride, rg ));

   DIPXJ( dip_ImageNew( &active, rg ));
   DIPXJ( dip_ImageCopyProperties( tmp, active ));
   DIPXJ( dip__ImageSetStride( active, stride ));
   DIPXJ( dip_ImageSetDataType( active, DIP_DT_BIN8 ));
   DIPXJ( dip_ImageForge( active ));

   DIPXJ( dip_ImageNew( &len1, rg ));
   DIPXJ( dip_ImageCopyProperties( tmp, len1 ));
   DIPXJ( dip__ImageSetStride( len1, stride ));
   DIPXJ( dip_ImageSetDataType( len1, DIP_DT_UINT16 ));
   DIPXJ( dip_ImageForge( len1 ));

   DIPXJ( dip_ImageNew( &len2, rg ));
   DIPXJ( dip_ImageCopyProperties( tmp, len2 ));
   DIPXJ( dip__ImageSetStride( len2, stride ));
   DIPXJ( dip_ImageSetDataType( len2, DIP_DT_UINT16 ));
   DIPXJ( dip_ImageForge( len2 ));

   if( constrained ) {

      DIPXJ( dip_ImageNew( &len3, rg ));
      DIPXJ( dip_ImageCopyProperties( tmp, len3 ));
      DIPXJ( dip__ImageSetStride( len3, stride ));
      DIPXJ( dip_ImageSetDataType( len3, DIP_DT_UINT16 ));
      DIPXJ( dip_ImageForge( len3 ));

      DIPXJ( dip_ImageNew( &len4, rg ));
      DIPXJ( dip_ImageCopyProperties( tmp, len4 ));
      DIPXJ( dip__ImageSetStride( len4, stride ));
      DIPXJ( dip_ImageSetDataType( len4, DIP_DT_UINT16 ));
      DIPXJ( dip_ImageForge( len4 ));

   }
   else {
      len3 = len4 = 0;
   }

   /* Get data pointers */
   DIPXJ( dip_ImageArrayNew( &imar, 6, rg));
   imar->array[0] = tmp;
   imar->array[1] = active;
   imar->array[2] = len1;
   imar->array[3] = len2;
   imar->array[4] = len3;
   imar->array[5] = len4;
   DIPXJ( dip_ImageGetData( imar, &vpar, 0, 0, 0, 0, 0, rg ));

   /* Create sorted indices array (skipping border) */
   if( mask ) {
      DIPXJ( DCPY__IndicesArrayCreateMask( mask, stride, &indices, &noInd, rg ));
   }
   else {
      DIPXJ( DCPY__IndicesArrayCreate( imsz, stride, &indices, &noInd, rg ));
   }
   DIPXJ( dip_SortIndices( vpar->array[ 0 ], indices, noInd, DIP_SORT_DEFAULT,
                           ovlDataType, DIP_DT_INT ));
   if( closing ) {
      /* Just invert the sorted indices -- why isn't this an option in dip_Sort ??? */
      dip_int tmp;
      for( ii=0; ii<noInd/2; ii++ ) {
         tmp = indices[ii];
         indices[ii] = indices[noInd-ii-1];
         indices[noInd-ii-1] = tmp;
      }
   }

   /* Arrays to hold the indices to the neighbours */
   nneigh = dipm_PowInt( 3, ndims-1 ); /* number of upstream neighbours identical to number downstream. */
   DIPXJ( dip_IntegerArrayNew( &indxup, nneigh, 0, rg ));
   DIPXJ( dip_IntegerArrayNew( &indxdn, nneigh, 0, rg ));

   /* Loop over all directions */
   /* ((3^ndims)-1)/2 directions */
   firstone = DIP_TRUE;
   DIPXJ( dip_IntegerArrayNew( &direction, ndims, -1, rg ));
   for (;;) {

      /* Check to see if this direction is "unique":
       * There must be at least one positive value, and the first non-negative value must be positive.
       */
      valid = DIP_FALSE;
      for( ii=0; ii<ndims; ii++ ) {
         if( direction->array[ii] != 0 ) {
            if( direction->array[ii] > 0 ) {
               valid = DIP_TRUE;
            }
            break;
         }
      }
      if( valid ) {

         /* Fill arrays with indices to neighbours */
         indxup->size = nneigh;
         indxdn->size = nneigh;
         DIPXJ( DCPY__MakeNeighbourLists( direction, stride, indxup, indxdn ));

         /* Initialise temporary images */
         if( !firstone ) {
            DIPXJ( dip_Copy( in, tmp ));
         }
         if( mask ) {
            DIPXJ( dip_Copy( mask, active ));
         }
         else {
            DIPXJ( dip_SetInteger( active, DIP__PO_ACTIVE, 0, DIP_FALSE ));
         }
         DIPXJ( dip_PlaneDoEdge( active, 0, 0 )); /* Set border pixels to inactive, we won't process them. */
         DIPXJ( dip_SetInteger ( len1, length, 0, DIP_FALSE ));
         DIPXJ( dip_SetInteger ( len2, length, 0, DIP_FALSE ));
         if( constrained ) {
            DIPXJ( dip_SetInteger ( len3, length, 0, DIP_FALSE ));
            DIPXJ( dip_SetInteger ( len4, length, 0, DIP_FALSE ));
         }

         /* Do the data-type-dependent thing */
         if( constrained ) {
#define DIP_OVL_ALLOW DIP_DTGID_REAL
#define DIP_OVL_FUNC DCPY__ConstrainedPathOpening
#define DIP_OVL_ARGS ( vpar->array[ 0 ], vpar->array[ 1 ], vpar->array[ 2 ], vpar->array[ 3 ], \
                       vpar->array[ 4 ], vpar->array[ 5 ], indices, noInd, indxup, indxdn, length )
#include "dip_ovl.h"
         }
         else {
#define DIP_OVL_ALLOW DIP_DTGID_REAL
#define DIP_OVL_FUNC DCPY__PathOpening
#define DIP_OVL_ARGS ( vpar->array[ 0 ], vpar->array[ 1 ], vpar->array[ 2 ], vpar->array[ 3 ], \
                       indices, noInd, indxup, indxdn, length )
#include "dip_ovl.h"
         }

         if( firstone ) {
            DIPXJ( dip_Copy( tmp, out ));
            firstone = DIP_FALSE;
         }
         else {
            if( closing ) {
               DIPXJ( dip_Min( tmp, out, out ));
            }
            else {
               DIPXJ( dip_Max( tmp, out, out ));
            }
         }

      }

      for( ii=0; ii<ndims; ii++ ) {
         direction->array[ii]++;
         if( direction->array[ii] <= 1 ) {
            break;
         }
         direction->array[ii] = -1;
      }
      if( ii == ndims ) {
         break;
      }
   }

dip_error:
   DIP_FNR_EXIT;
}


void mexFunction (int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]) {
   dip_Image in;
   dip_Image mask;
   dip_Image out;
   dip_int filterParam;
   dip_Boolean closing;
   dip_Boolean constrained;
   DML_ERROR_INIT;
   DML_CHK_NARGSIN (5);
   DML_CHK_NARGSOUT (1);
   DML_2DIP_IMAGE (0, in);
   DML_2DIP_MASKIMAGE (1, mask);
   DML_GENDIP_IMAGE (out);
   DML_2DIP_INT (2, filterParam);
   DML_2DIP_BOOLEAN (3, closing);
   DML_2DIP_BOOLEAN (4, constrained);
   DMLXJ ( DCPY_PathOpening (in, mask, out, filterParam, closing, constrained) );
   DML_2MEX_IMAGE (0, out);
dml_error:
   DML_ERROR_EXIT;
}
