/*
 * Filename: pathopening_commoncode.c
 *
 * Code common to both directedpathopening.c and pathopening.c.
 *
 * Author:
 * Cris Luengo, Centre for Image Analysis, Uppsala, Sweden.
 * Copyright 2008-2009.
 *
 * This and accompanying files contain code exactly as it is in DIPlib
 * for the functions dip_DirectedPathOpening() and dip_PathOpening().
 * Here we modified only the function names to avoid name conflicts
 * when linking against DIPlib.
 *
 * 5 December 2014: Replaced the calls to dip__PixelQueueIsEmpty with
 *    calls to dip_PixelQueueIsEmpty, as the former is not available
 *    outside of DIPlib. I hadn't noticed because Linux doesn't
 *    control exported library functions as strictly as Windows does.
 *
 * This code may be used freely for research purposes only. Do not
 * copy or distribute without the prior written consent by the author.
 *
 */

#ifndef DIP_TPI

#include "diplib.h"
#include "dip_developer.h"      /* for dip__ImageGetData() */
#include "dip_pixelqueue.h"     /* for dip_PixelQueue */
#include "pathopening.h"

static dip_Error DCPY__ConstrainedPropagateChanges
            ( dip_bin8*, dip_uint16*, dip_uint16*, dip_IntegerArray, dip_IntegerArray,
              dip_int, dip_PixelQueue, dip_PixelQueue );

static dip_Error DCPY__PropagateChanges
            ( dip_bin8*, dip_uint16*, dip_IntegerArray, dip_IntegerArray,
              dip_int, dip_PixelQueue, dip_PixelQueue );

/* Generate code for all different data types */
#define DIP_TPI_ALLOW DIP_DTGID_REAL
#define DIP_TPI_FILE "pathopening_commoncode.c"
#include "dip_tpi.h"


/*
 * Create indices array (skipping border)
 */
dip_Error DCPY__IndicesArrayCreate
(
   dip_IntegerArray imsz,
   dip_IntegerArray stride,
   dip_int* *array,
   dip_int* number,
   dip_Resources baserg
)
{
   DIP_FNR_DECLARE("DCPY__IndicesArrayCreate");
   dip_int noInd, ii, ptr, ndims;
   dip_IntegerArray coords;
   dip_int* indices;
   void* vptr;
   DIP_FNR_INITIALISE;

   ndims = imsz->size;
   DIPXJ( dip_IntegerArrayNew( &coords, ndims, 1, rg ));

   noInd = imsz->array[0] - 2;
   for (ii=1; ii<imsz->size; ii++) {
      noInd *= imsz->array[ii] - 2;
   }
   DIPXJ( dip_MemoryNew( &vptr, noInd*sizeof(dip_int), baserg ));
   indices = (dip_int*)vptr;

   for (;;) {
      ptr = 0;
      for (ii=0; ii<ndims; ii++) {
         ptr += coords->array[ii] * stride->array[ii];
      }
      for (ii=1; ii<imsz->array[0]-1; ii++) {
         *indices = ptr;
         indices++;
         ptr += stride->array[0];
      }
      for (ii=1; ii<ndims; ii++) {
         coords->array[ii]++;
         if (coords->array[ii] < imsz->array[ii]-1) {
            break;
         }
         coords->array[ii] = 1;
      }
      if (ii == ndims) {
         break;
      }
   }
   *array = (dip_int*)vptr;
   *number = noInd;

dip_error:
   DIP_FNR_EXIT;
}

/*
 * Create indices array from mask image (skipping border)
 */
dip_Error DCPY__IndicesArrayCreateMask
(
   dip_Image maskim,
   dip_IntegerArray stride,
   dip_int* *array,
   dip_int* number,
   dip_Resources baserg
)
{
   DIP_FNR_DECLARE("DCPY__IndicesArrayCreateMask");
   dip_int noInd, ii, ptr, mptr, ndims, plane;
   dip_IntegerArray coords, imsz, maskstride;
   dip_DataType dt;
   dip_int* indices;
   dip_bin8* mask;
   void* vptr;
   DIP_FNR_INITIALISE;

   DIPXJ( dip_IsScalar( maskim, 0 ));
   DIPXJ( dip_ImageGetDataType( maskim, &dt ));
   DIPTS( dt != DIP_DT_BIN8, "Mask image should be of data type dip_bin8." );

   DIPXJ( dip_ImageGetDimensions( maskim, &imsz, rg ));
   ndims = imsz->size;
   DIPTS( stride->size != ndims, "Stride array not same dimensionality as mask image." );
   DIPXJ( dip_ImageGetStride( maskim, &maskstride, rg ));
   DIPXJ( dip_ImageGetPlane( maskim, &plane ));
   plane = 1>>plane;
   DIPXJ( dip__ImageGetData( maskim, &vptr ));
   mask = (dip_bin8*)vptr;

   DIPXJ( dip_IntegerArrayNew( &coords, ndims, 1, rg ));
   noInd = 0;
   for (;;) {
      mptr = 0;
      for (ii=0; ii<ndims; ii++) {
         mptr += coords->array[ii] * maskstride->array[ii];
      }
      for (ii=1; ii<imsz->array[0]-1; ii++) {
         if (*(mask+mptr) & plane) {
            noInd++;
         }
         mptr += maskstride->array[0];
      }
      for (ii=1; ii<ndims; ii++) {
         coords->array[ii]++;
         if (coords->array[ii] < imsz->array[ii]-1) {
            break;
         }
         coords->array[ii] = 1;
      }
      if (ii == ndims) {
         break;
      }
   }

   DIPXJ( dip_MemoryNew( &vptr, noInd*sizeof(dip_int), baserg ));
   indices = (dip_int*)vptr;

   for (ii=0; ii<ndims; ii++) {
      coords->array[ii] = 1;
   }
   for (;;) {
      ptr = 0;
      mptr = 0;
      for (ii=0; ii<ndims; ii++) {
         ptr += coords->array[ii] * stride->array[ii];
         mptr += coords->array[ii] * maskstride->array[ii];
      }
      for (ii=1; ii<imsz->array[0]-1; ii++) {
         if (*(mask+mptr) & plane) {
            *indices = ptr;
            indices++;
         }
         ptr += stride->array[0];
         mptr += maskstride->array[0];
      }
      for (ii=1; ii<ndims; ii++) {
         coords->array[ii]++;
         if (coords->array[ii] < imsz->array[ii]-1) {
            break;
         }
         coords->array[ii] = 1;
      }
      if (ii == ndims) {
         break;
      }
   }
   *array = (dip_int*)vptr;
   *number = noInd;

dip_error:
   DIP_FNR_EXIT;
}


/*
 * Helper function that returns TRUE if 'coords' contains a direction that
 * is close to that of 'direction'. If 'coords' points at the origin, it
 * always returns FALSE. If 'coords' and 'direction' are identical, it also
 * returns FALSE.
 *
 * Of all the non-zero elements in 'direction', at least one must be identical
 * in 'coords', the other elements can differ by one. This defines a 90 degree
 * wedge in 2D, a 90 degree cone in 3D, and a similar set of angles in higher
 * dimensions.
 */
static dip_Boolean DCPY__IsValidNeighbour
(
   dip_IntegerArray direction,
   dip_IntegerArray coords
)
{
   dip_int ii, tmp;
   dip_Boolean isallzero = DIP_TRUE;
   dip_Boolean hasunchanged = DIP_FALSE;
   dip_Boolean isidentical = DIP_TRUE;
   for( ii=0; ii<coords->size; ii++ ) {
      isallzero = isallzero && ( coords->array[ii]==0 );
      isidentical = isidentical && ( coords->array[ii]==direction->array[ii] );
      if( direction->array[ii] != 0 ) {
         tmp = direction->array[ii] - coords->array[ii];
         tmp = DIP_ABS( tmp );
         if( tmp > 1 ) {
            return DIP_FALSE;
         }
         else if( tmp == 0 ) {
            hasunchanged = DIP_TRUE;
         }
      }
   }
   if( isidentical || isallzero || !hasunchanged ) {
      return DIP_FALSE;
   }
   else {
      return DIP_TRUE;
   }
}


/*
 * Helper function that creates a list of upstream and downstream neighbours
 * depending on 'direction', which is an array with elements -1, 0 and 1.
 */
dip_Error DCPY__MakeNeighbourLists
(
   dip_IntegerArray direction, /* input */
   dip_IntegerArray stride,    /* input */
   dip_IntegerArray indxup,    /* output, pre-allocated */
   dip_IntegerArray indxdn     /* output, pre-allocated */
)
{
   DIP_FNR_DECLARE("DCPY__MakeNeighbourLists");
   dip_int ndims;
   dip_int ii;
   dip_int ptrup, dist, nneigh;
   dip_IntegerArray coords;
   DIP_FNR_INITIALISE;

   ndims = direction->size;
   /* We put the neighbour given by "direction" first */
   ptrup = dist = 0;
   for( ii=0; ii<ndims; ii++ ) {
      ptrup += direction->array[ii] * stride->array[ii];
      dist += DIP_ABS(direction->array[ii]);
   }
   indxup->array[0] = ptrup;
   indxdn->array[0] = -ptrup;
   /* Now we list all other neighbours */
   nneigh = 1;
   DIPXJ( dip_IntegerArrayNew( &coords, ndims, -1, rg ));
   for (;;) {
      if( DCPY__IsValidNeighbour( direction, coords ) ) {
         DIPTS( nneigh >= indxup->size, "Assertion failed!" );
         ptrup = dist = 0;
         for( ii=0; ii<ndims; ii++ ) {
            ptrup += coords->array[ii] * stride->array[ii];
            dist += DIP_ABS(coords->array[ii]);
         }
         indxup->array[nneigh] = ptrup;
         indxdn->array[nneigh] = -ptrup;
         nneigh++;
      }
      for( ii=0; ii<ndims; ii++ ) {
         coords->array[ii]++;
         if( coords->array[ii] <= 1 ) {
            break;
         }
         coords->array[ii] = -1;
      }
      if( ii == ndims ) {
         break;
      }
   }
   indxup->size = nneigh;
   indxdn->size = nneigh;

dip_error:
   DIP_FNR_EXIT;
}


/*
 * Main functionality -- recursive updating of upstream (or downstream) neighbours.
 */
static dip_Error DCPY__ConstrainedPropagateChanges
(
   dip_bin8 *active,
   dip_uint16 *straight_length,
   dip_uint16 *other_length,     /* other_length >= straigh_length */
   dip_IntegerArray next,
   dip_IntegerArray prev,
   dip_int index,
   dip_PixelQueue queue,
   dip_PixelQueue changed
)
{
   DIP_FN_DECLARE("DCPY__ConstrainedPropagateChanges");
   dip_int ii, jj, len_s, len_o;
   dip_bin8 *aptr;
   dip_Boolean b;
   void* vptr;

   /* This pixel's length is 0 */
   straight_length[index] = 0;
   other_length[index] = 0;
   /* Enqueue the neighbours that are still active */
   for( jj=0; jj<next->size; jj++ ) {
      ii = index + next->array[jj];
      if( active[ii] & DIP__PO_ACTIVE ) {
         active[ii] |= DIP__PO_QUEUED;
         DIPXJ( dip_PixelQueuePush( queue, 0, (void*)&(active[ii]) ));
      }
   }
   /* Process pixels in queue */
   while( dip_PixelQueueIsEmpty( queue, &b ), !b ) {
      /* Pop a pixel */
      DIPXJ( dip_PixelQueuePop( queue, 0, &vptr, 0 ));
      aptr = (dip_bin8*)vptr;
      index = aptr-active;
      *aptr &= ~DIP__PO_QUEUED; /* Clear the queued bit */
      /* Update this pixel's lengths looking backwards */
      ii = index + prev->array[0];
      len_s = len_o = other_length[ii];
      for( jj=1; jj<prev->size; jj++ ) {
         ii = index + prev->array[jj];
         if( straight_length[ii] > len_o ) {
            len_o = straight_length[ii];
         }
      }
      len_s++;
      len_o++;
      if( len_s < straight_length[index] ) {
         straight_length[index] = len_s;
         /* Enqueue the neighbours that are still active */
         for( jj=0; jj<next->size; jj++ ) {
            ii = index + next->array[jj];
            if( (active[ii] & DIP__PO_ACTIVE) && !(active[ii] & DIP__PO_QUEUED) ) {
               active[ii] |= DIP__PO_QUEUED;
               DIPXJ( dip_PixelQueuePush( queue, 0, (void*)&(active[ii]) ));
            }
         }
         /* Put this one on the 'changed' list */
         if( !( *aptr & DIP__PO_CHANGED ) ) {
            *aptr |= DIP__PO_CHANGED; /* this is to make sure they're only pushed once */
            DIPXJ( dip_PixelQueuePush( changed, 0, (void*)aptr ));
         }
      }
      if( len_o < other_length[index] ) {
         other_length[index] = len_o;
         /* Enqueue the straight neighbour if still active */
         ii = index + next->array[0];
         if( (active[ii] & DIP__PO_ACTIVE) && !(active[ii] & DIP__PO_QUEUED) ) {
            active[ii] |= DIP__PO_QUEUED;
            DIPXJ( dip_PixelQueuePush( queue, 0, (void*)&(active[ii]) ));
         }
         /* Put this one on the 'changed' list */
         if( !( *aptr & DIP__PO_CHANGED ) ) {
            *aptr |= DIP__PO_CHANGED; /* this is to make sure they're only pushed once */
            DIPXJ( dip_PixelQueuePush( changed, 0, (void*)aptr ));
         }
      }
   }

dip_error:
   DIP_FN_EXIT;
}

static dip_Error DCPY__PropagateChanges
(
   dip_bin8 *active,
   dip_uint16 *length,
   dip_IntegerArray next,
   dip_IntegerArray prev,
   dip_int index,
   dip_PixelQueue queue,
   dip_PixelQueue changed
)
{
   DIP_FN_DECLARE("DCPY__PropagateChanges");
   dip_int ii, jj, len;
   dip_bin8 *aptr;
   dip_Boolean b;
   void* vptr;

   /* This pixel's length is 0 */
   length[index] = 0;
   /* Enqueue the neighbours that are still active */
   for( jj=0; jj<next->size; jj++ ) {
      ii = index + next->array[jj];
      if( active[ii] & DIP__PO_ACTIVE ) {
         active[ii] |= DIP__PO_QUEUED; /* this is to make sure they're only pushed once */
         DIPXJ( dip_PixelQueuePush( queue, 0, (void*)&(active[ii]) ));
      }
   }
   /* Process pixels in queue */
   while( dip_PixelQueueIsEmpty( queue, &b ), !b ) {
      /* Pop a pixel */
      DIPXJ( dip_PixelQueuePop( queue, 0, &vptr, 0 ));
      aptr = (dip_bin8*)vptr;
      index = aptr-active;
      *aptr &= ~DIP__PO_QUEUED; /* Clear the queued bit */
      /* Update this pixel's length looking backwards */
      ii = index + prev->array[0];
      len = length[ii];
      for( jj=1; jj<prev->size; jj++ ) {
         ii = index + prev->array[jj];
         if( length[ii] > len ) {
            len = length[ii];
         }
      }
      len++;
      if( len < length[index] ) {
         length[index] = len;
         /* Enqueue the neighbours that are still active */
         for( jj=0; jj<next->size; jj++ ) {
            ii = index + next->array[jj];
            if( (active[ii] & DIP__PO_ACTIVE) && !(active[ii] & DIP__PO_QUEUED) ) {
               active[ii] |= DIP__PO_QUEUED;
               DIPXJ( dip_PixelQueuePush( queue, 0, (void*)&(active[ii]) ));
            }
         }
         /* Put this one on the 'changed' list */
         if( !( *aptr & DIP__PO_CHANGED ) ) {
            *aptr |= DIP__PO_CHANGED; /* this is to make sure they're only pushed once */
            DIPXJ( dip_PixelQueuePush( changed, 0, (void*)aptr ));
         }
      }
   }

dip_error:
   DIP_FN_EXIT;
}

#else

/*
 * Data-type dependent portion of the code.
 * Iterate over all pixels of the image, in order of grey-value, and
 * update.
 */
DIP_TPI_DEFINE(DCPY__ConstrainedPathOpening)
(
   DIP_TPI *grey,               /* grey in & out */
   dip_bin8 *active,            /* temp: marks active pixels */
   dip_uint16 *slup,            /* temp: upstream length, straigh */
   dip_uint16 *olup,            /* temp: upstream length, non-straight */
   dip_uint16 *sldn,            /* temp: downstream length, straight */
   dip_uint16 *oldn,            /* temp: downstream length, non-straight */
   dip_int* indices,            /* array of length noInd, with indices into images */
   dip_int noInd,
   dip_IntegerArray indxup,     /* indices to upstream neighbours */
   dip_IntegerArray indxdn,     /* indices to upstream neighbours */
   dip_int length               /* param */
)
{
   DIP_FNR_DECLARE(DIP_TPI_NAME(DCPY__ConstrainedPathOpening));
   dip_int ii, jj, indx;
   dip_PixelQueue queue;
   dip_PixelQueue changed;
   dip_bin8 *aptr;
   dip_Boolean b;
   void *vptr;
   DIP_FNR_INITIALISE;

   DIPXJ( dip_PixelQueueNew( &queue,   0, indxup->size * length * length, rg ));
   DIPXJ( dip_PixelQueueNew( &changed, 0, indxup->size * length * length, rg ));

   for( jj=0; jj<noInd; jj++ ) {
      indx = indices[jj];
      if( !(active[indx] & DIP__PO_ACTIVE) ) {
         continue;
      }
      /* Propagate changes upstream */
      DIPXJ( DCPY__ConstrainedPropagateChanges( active, slup, olup, indxup, indxdn, indx, queue, changed ));
      /* Propagate changes downstream */
      DIPXJ( DCPY__ConstrainedPropagateChanges( active, sldn, oldn, indxdn, indxup, indx, queue, changed ));
      /* Go over changed pixels and update grey, active, etc. */
      while( dip_PixelQueueIsEmpty( changed, &b ), !b ) {
         DIPXJ( dip_PixelQueuePop( changed, 0, &vptr, 0 ));
         aptr = (dip_bin8*)vptr;
         *aptr &= ~DIP__PO_CHANGED;
         ii = aptr - active;
         if( ( slup[ii] + oldn[ii] - 1 < length ) &&
             ( olup[ii] + sldn[ii] - 1 < length )) {
            grey[ii] = grey[indx];
            active[ii] &= ~DIP__PO_ACTIVE;
            /*slup[ii] = 0;*/
            /*olup[ii] = 0;*/
            /*sldn[ii] = 0;*/
            /*oldn[ii] = 0;*/
         }
      }
      active[indx] &= ~DIP__PO_ACTIVE;
   }

dip_error:
   DIP_FNR_EXIT;
}

DIP_TPI_DEFINE(DCPY__PathOpening)
(
   DIP_TPI *grey,               /* grey in & out */
   dip_bin8 *active,            /* temp: marks active pixels */
   dip_uint16 *lup,             /* temp: upstream length */
   dip_uint16 *ldn,             /* temp: downstream length */
   dip_int* indices,            /* array of length noInd, with indices into images */
   dip_int noInd,
   dip_IntegerArray indxup,     /* indices to upstream neighbours */
   dip_IntegerArray indxdn,     /* indices to upstream neighbours */
   dip_int length               /* param */
)
{
   DIP_FNR_DECLARE(DIP_TPI_NAME(DCPY__PathOpening));
   dip_int ii, jj, indx;
   dip_PixelQueue queue, changed;
   dip_bin8 *aptr;
   dip_Boolean b;
   void *vptr;
   DIP_FNR_INITIALISE;

   DIPXJ( dip_PixelQueueNew( &queue,   0, indxup->size * length * length, rg ));
   DIPXJ( dip_PixelQueueNew( &changed, 0, indxup->size * length * length, rg ));

   for( jj=0; jj<noInd; jj++ ) {
      indx = indices[jj];
      if( !(active[indx] & DIP__PO_ACTIVE) ) {
         continue;
      }
      /* Propagate changes downstream */
      DIPXJ( DCPY__PropagateChanges( active, lup, indxup, indxdn, indx, queue, changed ));
      /* Propagate changes upstream */
      DIPXJ( DCPY__PropagateChanges( active, ldn, indxdn, indxup, indx, queue, changed ));
      /* Go over changed pixels and update grey, active, etc. */
      while( dip_PixelQueueIsEmpty( changed, &b ), !b ) {
         DIPXJ( dip_PixelQueuePop( changed, 0, &vptr, 0 ));
         aptr = (dip_bin8*)vptr;
         *aptr &= ~DIP__PO_CHANGED;
         ii = aptr - active;
         if( lup[ii] + ldn[ii] - 1 < length ) {
            grey[ii] = grey[indx];
            active[ii] &= ~DIP__PO_ACTIVE;
            /*lup[ii] = 0;*/
            /*ldn[ii] = 0;*/
         }
      }
      active[indx] &= ~DIP__PO_ACTIVE;
   }

dip_error:
   DIP_FNR_EXIT;
}

#endif
